/*
 * Definitions for multiple precision numbers
 * Assumptions are made that a FULL is twice the number of bits as a HALF,
 * and that numbers are stored in two's complement.
 *
 * The size of numbers is limited to 2^(21*16), which is over 100 digits.
 *
 * Copyright (c) 2014 by David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

typedef	unsigned short	HALF;
typedef	unsigned int	FULL;
typedef	void (*NUM_CATCHER)(const char * msg);


#define	HALFCOUNT	21
#define	HALFBITS	16
#define	FULLBITS	(HALFBITS * 2)
#define	FULLONE		((FULL) 1)
#define	HALFMAX		((HALF) ((FULLONE << HALFBITS) - 1))
#define	HALFTOPBIT	((HALF) (FULLONE << (HALFBITS - 1)))
#define	FULLTOPBIT	((FULL) (FULLONE << (FULLBITS - 1)))
#define	TOPHALF(n)	((HALF) (((FULL) (n)) >> HALFBITS))
#define	BOTHALF(n)	((HALF) ((FULL) (n)))
#define	BOTTOTOP(n)	(((FULL) (n)) << HALFBITS)


typedef	struct
{
	short	neg;			/* 1 if negative, 0 otherwise */
	short	size;			/* number of HALFs in number */
	HALF	value[HALFCOUNT];	/* value of number */
} NUMBER;


/*
 * Various quick tests.
 * These reference a number without making a copy of it.
 */
#define	isNonNeg(num)	((num).neg == 0)
#define	isNeg(num)	((num).neg != 0)
#define	isHalf(num)	((num).size == 1)
#define	isFull(num)	((num).size == 2)
#define	isZero(num)	(isHalf(num) && ((num).value[0] == 0))
#define	isUnit(num)	(isHalf(num) && ((num).value[0] == 1))
#define	isEven(num)	(((num).value[0] & 0x1) == 0)
#define	isOdd(num)	(((num).value[0] & 0x1) != 0)
#define	isOne(num)	(isNonNeg(num) && isUnit(num))
#define	isPos(num)	(isNonNeg(num) && !isZero(num))
#define	isNull(num)	((num).size == 0)
#define	isNotNull(num)	((num).size > 0)


/*
 * Test whether or not the number fits within a normal long value,
 * including the sign bit.
 */
#define	isSmall(num)	(isHalf(num) || \
			(isFull(num) && ((num).value[1] < HALFTOPBIT)))


/*
 * Defined numbers.
 */
extern	NUMBER	numZero;
extern	NUMBER	numOne;
extern	NUMBER	numNull;


/*
 * Numeric cache routines.
 */
extern	void	numInit(void);
extern	void	numReset(void);
extern	NUMBER * numStore(const NUMBER num);
extern	NUMBER * numStoreAddr(const NUMBER * num);


/*
 * Numeric routines.
 */
extern	NUM_CATCHER	numCatcher(NUM_CATCHER newfunc);

extern	NUMBER		numFromLong(long i);
extern	NUMBER		numAbs(NUMBER num);
extern	NUMBER		numNeg(NUMBER num);
extern	NUMBER		numInc(NUMBER num);
extern	NUMBER		numDec(NUMBER num);
extern	NUMBER		numAdd(NUMBER num1, NUMBER num2);
extern	NUMBER		numSub(NUMBER num1, NUMBER num2);
extern	NUMBER		numMin(NUMBER num1, NUMBER num2);
extern	NUMBER		numMax(NUMBER num1, NUMBER num2);

extern	void		numAbsAddr(NUMBER * num);
extern	void		numNegAddr(NUMBER * num);
extern	void		numIncAddr(NUMBER * num);
extern	void		numDecAddr(NUMBER * num);
extern	void		numAddAddr(NUMBER * num1, const NUMBER * num2);
extern	void		numAddAddrToAddr(NUMBER * ans, const NUMBER * num1, const NUMBER * num2);
extern	void		numSubAddr(NUMBER * num1, const NUMBER * num2);
extern	void		numSubAddrToAddr(NUMBER * ans, const NUMBER * num1, const NUMBER * num2);

extern	NUMBER		numMul(NUMBER num1, NUMBER num2);
extern	NUMBER		numMulLong(NUMBER num, long i);
extern	NUMBER		numAddLong(NUMBER num, long i);
extern	NUMBER		numAnd(NUMBER num1, NUMBER num2);
extern	NUMBER		numAndNot(NUMBER num1, NUMBER num2);
extern	NUMBER		numTwoPower(int pow);
extern	NUMBER		numDiv10(NUMBER num, int * rem);
extern	NUMBER		numHalf(NUMBER num);
extern	NUMBER		numScaleDown(NUMBER, int);

extern	long		numToLong(const NUMBER num);

extern	int		numCmp(const NUMBER num1, const NUMBER num2);
extern	int		numCmpAddr(const NUMBER * num1, const NUMBER * num2);
extern	int		numCmpAbsAddr(const NUMBER * num1, const NUMBER * num2);
extern	int		numCmpLong(const NUMBER num, long i);
extern	int		numCmpLongAddr(const NUMBER * num, long i);

extern	int		numLowBitNumber(NUMBER num);
extern	int		numHighBitNumber(NUMBER num);
extern	int		numBitCount(NUMBER num);

extern	const char *	numString(NUMBER num);

