/*
 * Read a file in pictoral format.
 *
 * Copyright (c) 2014 by David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "hashlife.h"


/*
 * Read the specified file name which contains a Life object in pictorial
 * format, and add it to the current view of the specified object.
 * Returns FALSE with a message printed if the file cannot be read or
 * is in the wrong format.
 */
BOOL
readPict(const char * name, OBJ * obj)
{
	FILE *	fp;
	NUMBER	num;
	NUMBER	curX;
	NUMBER	curY;
	long	line;
	int	ch;
	BOOL	sawNum;

	fp = fopen(name, "r");

	if (fp == 0)
	{
		fprintf(stderr, "Cannot open \"%s\" for reading: %s\n",
			name, strerror(errno));

		return FALSE;
	}

	line = 1;

	/*
	 * Look for sequences which are the following:
	 *	<number>.
	 *	<number>o
	 *	<number>\n
	 * which are OFF cells, ON cells, and newlines.
	 * Spaces and asterisks are also accepted for OFF and ON cells.
	 */
	curX = numZero;
	curY = numZero;
	sawNum = FALSE;

	while (TRUE)
	{
		ch = fgetc(fp);

		switch (ch)
		{
			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				if (!sawNum)
					num = numZero;

				sawNum = TRUE;
				num = numMulLong(num, 10);
				num = numAddLong(num, ch - '0');

				break;

			case 'O':
			case 'o':
			case '*':
				if (!sawNum)
					num = numOne;

				sawNum = FALSE;

				while (isPos(num))
				{
					setCell(obj, curX, curY);
					numDecAddr(&num);
					numIncAddr(&curX);
				}

				break;

			case '.':
			case ' ':
				if (!sawNum)
					num = numOne;

				sawNum = FALSE;

				numAddAddr(&curX, &num);

				break;

			case '\n':
				if (!sawNum)
					num = numOne;

				sawNum = FALSE;

				numAddAddr(&curY, &num);
				curX = numZero;

				line++;

				break;

			case EOF:
				if (sawNum)
					goto badFormat;

				if (ferror(fp))
					goto badEof;

				fclose(fp);

				return TRUE;

			default:
				goto badFormat;
		}
	}

	/*
	 * Error handling for illegal formats.
	 */
badFormat:
	if (ch == EOF)
		goto badEof;

	fprintf(stderr, "Bad format in \"%s\" at line %ld (character 0x%02x)\n",
		name, line, ch);

	fclose(fp);

	return FALSE;

badEof:
	if (ferror(fp))
	{
		fprintf(stderr, "Read error for \"%s\": %s\n",
			name, strerror(errno));

		fclose(fp);

		return FALSE;
	}

	fprintf(stderr, "Unexpected end of file for \"%s\"\n", name);
	fclose(fp);

	return FALSE;
}
