/*
 * Interface routines for X11 display.
 *
 * Copyright (c) 2014 by David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

#ifdef	X11

#include <stdio.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/StringDefs.h>

#include "hashlife.h"
#include "dev.h"


#define	WMRESX		20	/* X pixels to reserve for window manager */
#define	WMRESY		30	/* Y pixels to reserve for window manager */
#define	MINMAG		1	/* minimum magnification */
#define	MAXMAG		500	/* maximum magnification */
#define	DEFAULTMAG	20	/* default magnification */
#define	TEXTOFF		5	/* pixels for offsetting text from edges */


static	int	x11Open(DEV * dev);
static	void	x11Close(DEV * dev);
static	void	x11Update(DEV * dev, int inputFlag);
static	void	x11Refresh(DEV * dev);
static	int	x11ReadChar(DEV * dev, BOOL wait);
static	int	x11PeekInput(DEV * dev);
static	BOOL	x11CheckInput(DEV * dev);
static	int	x11SetMag(DEV * dev, int mag);
static	void	x11ShowStatus(DEV * dev, const char * str);
static	void	x11AddStatus(DEV * dev, const char * str);
static	void	x11BeginText(DEV * dev);
static	void	x11AddText(DEV * dev, const char * str);
static	void	x11EndText(DEV * dev);
static	void	x11ViewBegin(DEV * dev);
static	void	x11ViewEnd(DEV * dev);
static	void	x11ViewCells(DEV * dev, int count, int state);
static	void	x11ViewNextLine(DEV * dev);


DEV	x11Dev	=
{
	x11Open, x11Close, x11Update, x11Refresh, x11PeekInput, x11CheckInput,
	x11ReadChar, x11ViewBegin, x11ViewCells, x11ViewNextLine, x11ViewEnd,
	x11ShowStatus, x11AddStatus, x11BeginText, x11AddText, x11EndText,
	x11SetMag, 0, 0, 0, 0
};


/*
 * Other variables.
 */
static	Display *	display;	/* display for X */
static	int		screen;		/* screen number */
static	Window		rootWid;	/* top level window id */
static	Window		mainWid;	/* main window id */
static	Window		viewWid;	/* window id for view of cells */
static	Window		statWid;	/* window id for status line */
static	GC		statGC;		/* GC used for status line */
static	GC		viewGC;		/* GC used for view of cells */
static	XFontStruct *	font;		/* font for text */
static	Colormap	cmap;		/* default color map */
static	long		black;		/* black pixel value */
static	long		white;		/* white pixel value */
static	long		red;		/* red pixel value */
static	long		blue;		/* blue pixel value */
static	long		grey;		/* grey pixel value */
static	int		charHeight;	/* height of characters */
static	int		charAscent;	/* ascent of characters */
static	int		spaceWidth;	/* width of a space */
static	int		statWidth;	/* width of status window */
static	int		statHeight;	/* height of status window */
static	int		viewWidth;	/* width of view window */
static	int		viewHeight;	/* height of view window */
static	int		usedStatus;	/* space used in status line so far */
static	int		textX;		/* x position for new text */
static	int		textY;		/* y position for new text */
static	int		textCol;	/* column number for new text */
static	int		viewRow;	/* current viewing row */
static	int		viewCol;	/* current viewing column */
static	int		viewMag;	/* magnification scale factor */
static	int		savedChar;	/* saved character */
static	BOOL		haveSavedChar;	/* TRUE if have a saved character */


/*
 * Local procedures.
 */
static	int	doEvent(XEvent * event);
static	int	doExposure(XExposeEvent * ep);
static	int	doButtonDown(XButtonEvent * bp);
static	int	doKeyDown(XKeyEvent * kp);


/*
 * Open the graphics device.
 */
static int
x11Open(DEV * dev)
{
	XColor		cellColor;
	XColor		nameColor;
	XSizeHints	hints;
	char *		dummyArgv[1];

	display = XOpenDisplay(0);

	if (display == 0)
	{
		fprintf(stderr, "Cannot open graphics\n");

		return -1;
	}

	screen = DefaultScreen(display);

	font = XLoadQueryFont(display, "fixed");

	if (font == 0)
	{
		fprintf(stderr, "Cannot load font\n");

		return -1;
	}

	charAscent = font->ascent;
	charHeight = charAscent + font->descent;
	spaceWidth = font->max_bounds.width;

	/*
	 * Allocate colors, defaulting them if they are not available.
	 */
	black = BlackPixel(display, screen);
	white = WhitePixel(display, screen);
	cmap = DefaultColormap(display, screen);

	red = white;
	blue = black;
	grey = black;

	if (XAllocNamedColor(display, cmap, "red", &cellColor, &nameColor))
	{
		if (cellColor.pixel != black)
			red = cellColor.pixel;
	}

	if (XAllocNamedColor(display, cmap, "blue", &cellColor, &nameColor))
	{
		if (cellColor.pixel != white)
			blue = cellColor.pixel;
	}

	if (XAllocNamedColor(display, cmap, "grey", &cellColor, &nameColor))
	{
		if (cellColor.pixel != white)
			grey = cellColor.pixel;
	}

	viewWidth = DisplayWidth(display, screen) - WMRESX;
	statWidth = viewWidth;
	statHeight = charHeight + TEXTOFF * 2;
	viewHeight = DisplayHeight(display, screen) - statHeight - WMRESY;

	rootWid = RootWindow(display, screen);

	mainWid = XCreateSimpleWindow(display, rootWid,
		0, 0, viewWidth, viewHeight + statHeight,
		0, 0, 0);

	statWid = XCreateSimpleWindow(display, mainWid,
		0, 0, viewWidth, statHeight,
		0, 0, grey);

	viewWid = XCreateSimpleWindow(display, mainWid,
		0, statHeight, viewWidth, viewHeight,
		0, 0, blue);

	XSelectInput(display, statWid, ExposureMask | KeyPressMask);

	XSelectInput(display, viewWid, ExposureMask | KeyPressMask |
		ButtonPressMask);

	hints.flags = PSize | PMinSize | PMaxSize;
	hints.width = viewWidth;
	hints.min_width = hints.width;
	hints.max_width = hints.width;
	hints.height = viewHeight + statHeight;
	hints.min_height = hints.height;
	hints.max_height = hints.height;

	XSetStandardProperties(display, mainWid, "life", "life", None,
		dummyArgv, 0, &hints);

	XMapWindow(display, mainWid);
	XMapWindow(display, statWid);
	XMapWindow(display, viewWid);

	statGC = XCreateGC(display, statWid, 0, 0);
	viewGC = XCreateGC(display, viewWid, 0, 0);

	XSetForeground(display, viewGC, white);
	XSetForeground(display, statGC, blue);
	XSetFont(display, statGC, font->fid);

	viewMag = DEFAULTMAG;

	dev->rows = viewHeight / viewMag;
	dev->cols = viewWidth / viewMag;
	dev->textRows = (viewHeight - TEXTOFF * 2) / charHeight;
	dev->textCols = (viewWidth - TEXTOFF * 2) / spaceWidth;

	haveSavedChar = FALSE;

	return 0;
}


static void
x11Close(DEV * dev)
{
	if (display)
		XCloseDisplay(display);
}


static void
x11Update(DEV * dev, int inputFlag)
{
	XSync(display, False);
}


static void
x11Refresh(DEV * dev)
{
	XSync(display, False);
}


/*
 * Read the next character from the terminal, waiting if requested.
 * If not waiting and no character is ready, then -1 is returned.
 * Returns characters from a generated string first, if present.
 */
static int
x11ReadChar(DEV * dev, BOOL wait)
{
	XEvent	event;
	int	ch;

	if (haveSavedChar)
	{
		haveSavedChar = FALSE;

		return savedChar;
	}

	do
	{
		if (!wait && (XEventsQueued(display, QueuedAfterFlush) <= 0))
			return -1;

		XNextEvent(display, &event);

		ch = doEvent(&event);
	}
	while (ch < 0);

	return ch;
}


/*
 * Peek at the next input character.
 */
static int
x11PeekInput(DEV * dev)
{
	int	ch;

	ch = x11ReadChar(dev, TRUE);

	if (ch < 0)
		return -1;

	savedChar = ch;
	haveSavedChar = TRUE;

	return ch;
}


/*
 * See if input is ready from the terminal.
 * Returns FALSE if not, TRUE if so.
 */
static BOOL
x11CheckInput(DEV * dev)
{
	XEvent	event;

	if (haveSavedChar)
		return TRUE;

	while (TRUE)
	{
		if (XEventsQueued(display, QueuedAfterFlush) <= 0)
			return FALSE;

		XPeekEvent(display, &event);

		if (event.type == KeyPress)
			return TRUE;

		XNextEvent(display, &event);
		doEvent(&event);
	}
}


/*
 * Set the magnification factor.
 */
static int
x11SetMag(DEV * dev, int mag)
{
	if ((mag < MINMAG) || (mag > MAXMAG))
		return -1;

	viewMag = mag;

	dev->rows = viewHeight / viewMag;
	dev->cols = viewWidth / viewMag;

	return 0;
}


/*
 * Update the status line.
 */
static void
x11ShowStatus(DEV * dev, const char * str)
{
	int		len;
	int		width;

	len = strlen(str);
	XClearWindow(display, statWid);
	width = XTextWidth(font, str, len);

	XDrawString(display, statWid, statGC, TEXTOFF,
		charAscent + TEXTOFF, str, len);

	usedStatus = width + TEXTOFF;
}


/*
 * Add to the status line.
 * The assumption is made that this routine is used only for input,
 * and thus the text is ended with an underscore as a prompt.
 */
static void
x11AddStatus(DEV * dev, const char * str)
{
	int		len;
	static	int	usWidth;

	len = strlen(str);

	if (usWidth == 0)
		usWidth = XTextWidth(font, "_", 1);

	XClearArea(display, statWid, usedStatus, 0, usWidth, 0, False);

	XDrawString(display, statWid, statGC, usedStatus, charAscent + TEXTOFF,
		str, len);

	usedStatus += XTextWidth(font, str, len);

	XDrawString(display, statWid, statGC, usedStatus, charAscent + TEXTOFF,
		"_", 1);
}


/*
 * Begin to show text information.
 * We use the view window for this.
 */
static void
x11BeginText(DEV * dev)
{
	XClearWindow(display, viewWid);

	textCol = 0;
	textX = TEXTOFF;
	textY = charAscent + TEXTOFF;
}


/*
 * Add information to the text display.
 * We must interpret certain control codes.
 */
static void
x11AddText(DEV * dev, const char * str)
{
	const char *	cp;
	int		len;

	cp = str;

	for (;;)
	{
		while (*cp >= ' ')
			cp++;

		len = cp - str;

		if (len > 0)
		{
			XDrawString(display, viewWid, viewGC, textX, textY,
				str, len);

			textCol += len;
			textX += XTextWidth(font, str, len);
		}

		switch (*cp++)
		{
			case '\0':
				return;

			case '\n':
				textCol = 0;
				textX = TEXTOFF;
				textY += charHeight;

				break;

			case '\t':
				len = 8 - (textCol + 1) % 8;

				XDrawString(display, viewWid, viewGC,
					textX, textY, "        ", len);

				textCol += len;
				textX += spaceWidth * len;

				break;
		}

		str = cp;
	}
}


/*
 * End the text display.
 */
static void
x11EndText(DEV * dev)
{
}


/*
 * Handle the specified event.
 * Returns a character generated by the event, or else -1 if
 * the event does not produce a character.
 */
static int
doEvent(XEvent * event)
{
	switch (event->type)
	{
		case Expose:
			return doExposure(&event->xexpose);

		case ButtonPress:
			return doButtonDown(&event->xbutton);

		case KeyPress:
			return doKeyDown(&event->xkey);
	}

	return -1;
}


/*
 * Handle an exposure event.
 * This also consumes all other pending exposure events.
 */
static int
doExposure(XExposeEvent * ep)
{
	int	i;
	XEvent	event;

	for (i = 0; i < ep->count; i++)
		XNextEvent(display, &event);

	showView(viewObj);
	showStatus(viewObj);

	return -1;
}


/*
 * Handle a button down event.
 * This currently has no use.
 */
static int
doButtonDown(XButtonEvent * bp)
{
	return -1;
}


/*
 * Return a key typed in our window, or -1 if no such key was typed.
 */
static int
doKeyDown(XKeyEvent * kp)
{
	KeySym	keySym;
	int	len;
	int	ch;
	char	buf[12];

	if ((kp->window != statWid) && (kp->window != viewWid))
		return -1;

	len = XLookupString(kp, buf, sizeof(buf), &keySym, 0);

	if ((len != 1) || IsModifierKey(keySym))
		return -1;

	ch = buf[0];

	if (ch == '\r')
		ch = '\n';

	return ch;
}


/*
 * Start viewing for cells.
 */
static void
x11ViewBegin(DEV * dev)
{
	int i;

	XClearWindow(display, viewWid);

	viewRow = 0;
	viewCol = 0;

	/*
	 * Draw the grid if the magnification is large enough.
	 */
	if (viewMag < 10)
		return;

	for (i = 0; i <= viewHeight; i += viewMag)
	{
		XDrawLine(display, viewWid, viewGC, 0, i, viewWidth, i);
	}

	for (i = 0; i <= viewWidth; i += viewMag)
	{
		XDrawLine(display, viewWid, viewGC, i, 0, i, viewHeight);
	}

	XFlush(display);
}


/*
 * Terminate viewing of cells.
 */
static void
x11ViewEnd(DEV * dev)
{
	XFlush(display);
}


/*
 * Show a number of cells on the current line.
 */
static void
x11ViewCells(DEV * dev, int count, int state)
{
	if (state == 0)
	{
		viewCol += count;

		return;
	}

	if (viewMag == 1)
	{
		if (count < 5)
		{
			while (count-- > 0)
				XDrawPoint(display, viewWid, viewGC, viewCol++, viewRow);

			return;
		}

		XDrawLine(display, viewWid, viewGC, viewCol, viewRow, 
			viewCol + count - 1, viewRow);

		viewCol += count;

		return;
	}
	

	/*
	 * Here if the viewing magnification is larger than one.
	 */
	while (count-- > 0)
	{
		XFillRectangle(display, viewWid, viewGC,
			viewCol * viewMag, viewRow * viewMag,
			viewMag - 1, viewMag - 1);

		viewCol++;
	}
}


/*
 * Move the view to the next line.
 */
static void
x11ViewNextLine(DEV * dev)
{
	viewRow++;
	viewCol = 0;
}

#endif
